#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <libgen.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <glib.h>
#include <math.h>
#include <stdbool.h>
#include <syslog.h>
#include <errno.h>
#include <gio/gdesktopappinfo.h>

#include "unified-conf.h"

int create_directory_recursively(const char *path) // 递归创建不存在的目录
{
    char tmp[256];
    char *p = NULL;
    size_t len;

    snprintf(tmp, sizeof(tmp), "%s", path);
    len = strlen(tmp);
    if (tmp[len - 1] == '/')
        tmp[len - 1] = 0;
    for (p = tmp + 1; *p; p++)
        if (*p == '/')
        {
            *p = 0;
            mkdir(tmp, 0755);
            *p = '/';
        }
    return mkdir(tmp, 0755);
}

bool env_target_file_has_value(char *target_line) // 判断目标文件内是否已经包含了该key、value行
{
    FILE *fp = fopen(ENV_FILE_TARGET, "r");
    char buffer[1024];
    while (fgets(buffer, sizeof(buffer), fp))
    {
        // 去掉可能的换行符
        buffer[strcspn(buffer, "\n")] = 0;
        if (strcmp(buffer, target_line) == 0) // 比较当前行与目标字符串
        {
            syslog(LOG_INFO,"This line (%s) has already been included in the file",target_line);
            fclose(fp);
            return true; // 找到匹配的行
        }
    }
    fclose(fp);
    return false;
}

void process_conf_env(GKeyFile *config)
{
    // 处理env的目标文件
    FILE *target_fp;
    gsize length;
    GError *error = NULL;
    // 打开目标文件，如果不存在则创建
    if (create_directory_recursively(ENV_DIR_TARGET) == 0 || errno == EEXIST)
    {
        target_fp = fopen(ENV_FILE_TARGET, "a+"); // 使用"a"模式打开以追加内容

        if (!target_fp)
        {
            syslog(LOG_ERR,"Fopen target file : %s",error->message);
            g_error_free(error);
            return;
        }

        gchar **keys = g_key_file_get_keys(config, GROUP_NAME_ENV, NULL, &error);
        if (error)
        {
            syslog(LOG_ERR,"Error getting keys: %s",error->message);
            g_error_free(error);
            fclose(target_fp);
            return; // 清理并退出...
        }
        length = g_strv_length(keys);
        for (int i = 0; i < length; i++)
        {
            gchar *value = g_key_file_get_value(config, GROUP_NAME_ENV, keys[i], &error);
            char target_line[1024] = {0};
            sprintf(target_line, "%s=%s", keys[i], value);
            if (!env_target_file_has_value(target_line))
            {
                syslog(LOG_INFO,"Add this line (%s) to the configuration file",target_line);
                fprintf(target_fp, "%s=%s\n", keys[i], value);
            }
        }
        // 释放数组内存方法
        g_strfreev(keys);
        fclose(target_fp);
        g_error_free(error);
    }
}

/*源文件格式：SYSTEMD_UNIT_PATH=/var/opt/kare/lib/systemd/system-test:
目标文件格式：ManagerEnvironment=SYSTEMD_UNIT_PATH=/var/opt/kare/lib/systemd/system:
即源文件的key与value均为目标文件的value
*/
void update_manager_environment(const char *conf_dir,const char *conf_file,const char *update_key, const char *update_value)
{
    GError *error = NULL;
    char environment_value[1024] = {0}; // 用于存放修改后的value值
    char *token;
    char *saveptr;
    char *key, *value;
    bool flag = false; // 未找到key
    // 判断服务的目标文件是否存在，不存在则创建文件、再判断组是否存在
    if (create_directory_recursively(conf_dir) == 0 || errno == EEXIST)
    {
        if (access(conf_file, F_OK) != 0)
            creat(conf_file, 0666);

        GKeyFile *config = g_key_file_new(); // 创建目标文件对应的GKeyFile
        g_key_file_load_from_file(config, conf_file, G_KEY_FILE_KEEP_COMMENTS | G_KEY_FILE_KEEP_TRANSLATIONS, NULL);
        if (!g_key_file_has_group(config, GROUP_MAN_SER) || !g_key_file_has_key(config, GROUP_MAN_SER, KEY_MAN_ENV_SER, NULL)) // manager组或ManagerEnvironment不存在，则直接设置key与value,组与key会被自动创建
        {
            syslog(LOG_INFO,"Without Manager or ManagerEnvironment,Add key(%s) and value(%s) to it ",update_key,update_value);
            sprintf(environment_value, "%s=%s", update_key, update_value);                    // 构造目标文件的value

            g_key_file_set_string(config, GROUP_MAN_SER, KEY_MAN_ENV_SER, environment_value); // 设置目标文件的value
            if (g_key_file_save_to_file(config, conf_file, &error))                     // 保存至目标文件
            {
                if (error)
                    syslog(LOG_ERR,"Error save file : %s",error->message);
            }
            g_key_file_free(config);
            g_error_free(error);
            return;
        }

        gchar *manager_environment = g_key_file_get_string(config, GROUP_MAN_SER, KEY_MAN_ENV_SER, NULL);
        // 使用空格作为分隔符来分割字符串
        token = strtok_r(manager_environment, " ", &saveptr);

        while (token != NULL)
        {
            // 使用'='作为分隔符来分割key和value
            char *eq = strchr(token, '=');
            if (eq != NULL)
            {
                *eq = '\0'; // 将'='替换为字符串结束符
                key = token;
                value = eq + 1; // 跳过'='字符，指向value的开始
                sprintf(environment_value, "%s=", key);
                if (strcmp(key, update_key) == 0 && strstr(value, update_value) == NULL) // 找到对应的key且原value不包含update_value
                {
                    // 将update_value的值补充到key=后面且不影响原字符串其他部分内容
                    strcat(environment_value, update_value);
                    flag = true;
                }
                strcat(environment_value, value);
            }

            // 获取下一个token
            token = strtok_r(NULL, " ", &saveptr);
            strcat(environment_value, " ");
        }

        if (flag == false)
        {
            sprintf(environment_value, "%s%s%s ", key, "=", update_value);
        }

        g_key_file_set_string(config, GROUP_MAN_SER, KEY_MAN_ENV_SER, environment_value);
        g_key_file_save_to_file(config, conf_file, &error);
        if (error)
            syslog(LOG_ERR,"Error save file : %s",error->message);
        g_key_file_free(config);
        g_error_free(error);
    }
}

void process_conf_service(GKeyFile *source_config)
{
    gsize length;
    GError *error = NULL;
    gchar **source_keys = g_key_file_get_keys(source_config, GROUP_NAME_SER, NULL, &error); // 读取源文件的key列表
    if (error)
    {
        syslog(LOG_ERR,"Error getting keys: %s",error->message);
        g_error_free(error);
        return; // 清理并退出...
    }
    length = g_strv_length(source_keys);
    for (int i = 0; i < length; i++)
    {
        gchar *source_value = g_key_file_get_value(source_config, GROUP_NAME_SER, source_keys[i], &error); ////source_keys[i]、source_value均为源文件内的key与value
        update_manager_environment(SER_USR_TARGET_DIR,SER_USR_TARGET_FILE,source_keys[i], source_value);
        update_manager_environment(SER_SYS_TARGET_DIR,SER_SYS_TARGET_FILE,source_keys[i], source_value);
    }
    // 释放数组内存方法
    g_strfreev(source_keys);
    g_error_free(error);
}

void process_conf(const char *filename)
{
    GError *error = NULL;
    gsize length;
    GKeyFile *config = g_key_file_new();
    g_key_file_load_from_file(config, filename, G_KEY_FILE_KEEP_COMMENTS | G_KEY_FILE_KEEP_TRANSLATIONS, NULL);

    // 处理env函数
    process_conf_env(config);
    // 处理service函数
    process_conf_service(config);

    g_key_file_free(config);
    g_error_free(error);
}

void parse_conf_file()
{
    DIR *dir;
    struct dirent *entry;

    // 打开源目录
    dir = opendir(SOURCE_DIR);
    if (!dir)
    {
        perror("opendir");
        return;
    }

    // 遍历目录
    while ((entry = readdir(dir)) != NULL)
    {
        char *filename = entry->d_name;
        if (strstr(filename, ".conf") && strcmp(filename, ".") != 0 && strcmp(filename, "..") != 0)
        {
            char full_path[1024];
            snprintf(full_path, sizeof(full_path), "%s%s", SOURCE_DIR, filename);
            process_conf(full_path);
        }
    }
    closedir(dir);
    return;
}
